/*
 * Program: MQWhat
 *
 * Output MQ information that is contained in the Windows registry.
 *
 * @author Roger Lacroix, Capitalware Inc.
 * @return 0 for ok
 * @version 1.0.0
 * @license Apache 2 License
 */
#define PROGRAM_NAME     "MQWhat"
#define PROGRAM_VERSION  "1.2.0"
#define AUTHOR_NAME      "Capitalware Inc."
static const char * EYECATCHER_EyeCatcher  = PROGRAM_NAME "  ";
static const char * EYECATCHER_Version     = PROGRAM_VERSION "  ";
static const char * EYECATCHER_Author      = AUTHOR_NAME "  ";
static const char * EYECATCHER_Datestamp   = __DATE__ ;
static const char * EYECATCHER_Timestamp   = __TIME__ ;
static const char * EYECATCHER_Filename    = __FILE__ ;

/*
 * Headers
 */
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <stdio.h>

/* MQ headers */
#include <cmqc.h>

/*
 * MQWhat header file
 */
#include "MQWhat.h"

/*
 * Global Variables
 */
QMGR_T QMgrList[50];

HANDLE hChildStd_IN_Rd = NULL;
HANDLE hChildStd_IN_Wr = NULL;
HANDLE hChildStd_OUT_Rd = NULL;
HANDLE hChildStd_OUT_Wr = NULL;

/*
 * Function Name
 *  Mainline
 *
 * Description
 *  This is the main entry point to the program..
 *
 * Input parameters
 *  N/A
 *
 * Output
 *  MQ reistry information to the user's screen
 *
 * Return Value
 *  0
 */
int main(int argc, char **argv)
{
   printf("%s v%s by %s\n", PROGRAM_NAME, PROGRAM_VERSION, AUTHOR_NAME);
   printf("(c) Copyright %s 2009. All rights reserved.\n", AUTHOR_NAME);

   memset(&QMgrList, '\0', sizeof(QMgrList));
   GetMQVersion();
   GetMQComponentList();
   GetMQMaintenanceList();
   GetMQQMgrsList();
   GetMQIndividualQMgrInfo();

   return 0;
}

/*
 * Function Name
 *  GetMQVersion
 *
 * Description
 *  This function will output all MQ 'version' related information.
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQVersion()
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;
   DWORD dwSize = MAX_PATH;
   OSVERSIONINFOEX verInfo;    // Struct for version info
   char osName[128] = "";      // default to nothing

   ZeroMemory(&verInfo, sizeof(OSVERSIONINFO));
   verInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

   strcpy( lpszSubKey, SUBKEY_MQ_VERSION_SETTING );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_QUERY_VALUE,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      printf("+-----------------------------------------------------+\n");
      printf("| Currently Installed Version Information             |\n");
      printf("+-----------------------------------------------------+\n");
      DumpKey( "ProductType", hkCurKey, "", "\t");
      DumpKey( "VRMF", hkCurKey, "", "\t\t");
      DumpKey_DWord( "MQServerVersion", hkCurKey, "\t");
      DumpKey_DWord( "MQServerRelease", hkCurKey, "\t");
      DumpKey( "GSKitPackaged", hkCurKey, "", "\t");
      DumpKey( "FilePath", hkCurKey, "", "\t");

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQVersion : RegOpenKeyEx : lRet=%d\n", lRet);

   // Get major and minor number
   if (GetVersionEx((OSVERSIONINFO *) &verInfo))
   {
      switch (verInfo.dwPlatformId)
      {
         case VER_PLATFORM_WIN32_WINDOWS:
         {
            if ( (verInfo.dwMajorVersion == 4) && (verInfo.dwMinorVersion == 0) )
               strcpy(osName, "Windows 95");
            else if ( (verInfo.dwMajorVersion == 4) && (verInfo.dwMinorVersion == 10) )
               strcpy(osName, "Windows 98");
            else if ( (verInfo.dwMajorVersion == 4) && (verInfo.dwMinorVersion == 90) )
               strcpy(osName, "Windows ME");
            break;
         }

         case VER_PLATFORM_WIN32_NT:
         {
            if ( (verInfo.dwMajorVersion == 4) && (verInfo.dwMinorVersion == 0) )
               strcpy(osName, "Windows NT");
            else if ( (verInfo.dwMajorVersion == 5) && (verInfo.dwMinorVersion == 0) )
               strcpy(osName, "Windows 2000");
            else if ( (verInfo.dwMajorVersion == 5) && (verInfo.dwMinorVersion == 1) )
               strcpy(osName, "Windows XP");
            else if ( (verInfo.dwMajorVersion == 5) && (verInfo.dwMinorVersion == 2) )
            {
               if( GetSystemMetrics(SM_SERVERR2) )
                  strcpy(osName, "Windows Server 2003");
               else
                  strcpy(osName, "Windows Server 2003 R2");
            }
            else if ( (verInfo.dwMajorVersion == 6) && (verInfo.dwMinorVersion == 0) )
               strcpy(osName, "Windows Vista");
            else if ( (verInfo.dwMajorVersion == 6) && (verInfo.dwMinorVersion == 1) )
               strcpy(osName, "Windows 7");
            break;
         }

         default:
         {
            if ( (verInfo.dwMajorVersion == 6) && (verInfo.dwMinorVersion == 0) )
               strcpy(osName, "Windows Vista");
            else if ( (verInfo.dwMajorVersion == 6) && (verInfo.dwMinorVersion == 1) )
               strcpy(osName, "Windows 7");
            else
               printf("Error: Unknown Windows operating system.\n");
            break;
         }
      }

      printf("Operating System= %s %s (Build %d)\n",
             osName,
             verInfo.szCSDVersion,
             verInfo.dwBuildNumber & 0xFFFF);
   }
   else
      printf("Error: Could not get OS Version info\n");
}

/*
 * Function Name
 *  GetMQComponentList
 *
 * Description
 *  This function will output all MQ 'component' related information.
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQComponentList()
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;
   DWORD dwSize = MAX_PATH;
   int   KeyIndex=0;
   char  KeyName[MAX_PATH];
   char  Tabs[32];

   strcpy( lpszSubKey, SUBKEY_MQ_COMPONENT_SETTING );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_READ,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      printf("+-----------------------------------------------------+\n");
      printf("| Currently Installed Components                      |\n");
      printf("+-----------------------------------------------------+\n");

      /* Enumerate all of the descendents of this child. */
      while (RegEnumValue( hkCurKey,
                           KeyIndex,
                           KeyName,
                           &dwSize,
                           NULL,
                           NULL,
                           NULL,
                           NULL) == ERROR_SUCCESS)
      {
         if (strlen(KeyName) > 0)
         {
            CreateTabs(KeyName, Tabs, sizeof(Tabs), 4);
            DumpKey( KeyName, hkCurKey, "", Tabs);
         }
         KeyIndex++;
         dwSize = sizeof(KeyName);
      }

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQComponentList : RegOpenKeyEx : lRet=%d\n", lRet);
}

/*
 * Function Name
 *  GetMQMaintenanceList
 *
 * Description
 *  This function will output all MQ 'maintenance' related information.
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQMaintenanceList()
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;
   DWORD dwSize = MAX_PATH;
   int   KeyIndex=0;
   char  KeyName[MAX_PATH];

   strcpy( lpszSubKey, SUBKEY_MQ_MAINT_SETTING );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_READ,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      printf("+-----------------------------------------------------+\n");
      printf("| Currently Applied Maintenance                       |\n");
      printf("+-----------------------------------------------------+\n");
      DumpKey( "Shipped", hkCurKey, "", "\t");

      /* Enumerate all of the descendents of this child. */
      while (RegEnumValue( hkCurKey,
                           KeyIndex,
                           KeyName,
                           &dwSize,
                           NULL,
                           NULL,
                           NULL,
                           NULL) == ERROR_SUCCESS)
      {
         if (strlen(KeyName) > 0)
         {
            // Skip "Shipped" key as we already did it.
            if (strcmp(KeyName, "Shipped") != 0)
               printf("%s\n", KeyName);
         }
         KeyIndex++;
         dwSize = sizeof(KeyName);
      }

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQMaintenanceList : RegOpenKeyEx : lRet=%d\n", lRet);
}

/*
 * Function Name
 *  GetMQQMgrsList
 *
 * Description
 *  This function will output the names of all created MQ queue managers
 *  and their current status.
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQQMgrsList()
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;
   DWORD dwSize = MAX_PATH;
   char  QMgrName[MQ_Q_MGR_NAME_LENGTH + 1];
   char  ConvertedName[MQ_Q_MGR_NAME_LENGTH + 1];
   char  DefaultQMgrName[MQ_Q_MGR_NAME_LENGTH + 1];
   char  Status[MAX_PATH];
   int   KeyIndex=0;
   char  Tabs[32];
   QMGR_T *qq = QMgrList;

   GetMQDefaultQMgr(DefaultQMgrName, sizeof(DefaultQMgrName));

   strcpy( lpszSubKey, SUBKEY_MQ_QMGRS_SETTING );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_READ,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      printf("+-----------------------------------------------------+\n");
      printf("| Queue Managers                                      |\n");
      printf("+-----------------------------------------------------+\n");

      printf("Default\tQMgr Name\t\tStatus\n");
      printf("-------\t---------\t\t------\n");

      /* Enumerate all of the descendents of this child. */
      while (RegEnumKeyEx( hkCurKey,
                           KeyIndex,
                           QMgrName,
                           &dwSize,
                           NULL,
                           NULL,
                           NULL,
                           NULL) == ERROR_SUCCESS)
      {
         CreateTabs(QMgrName, Tabs, sizeof(Tabs), 4);
         strncpy(qq->Name, QMgrName, MQ_Q_MGR_NAME_LENGTH);
         GetMQQMgrStatus(QMgrName, Status, sizeof(QMgrName));
         ConvertMQObjectName(ConvertedName, QMgrName);
         if (strcmp(ConvertedName, DefaultQMgrName) == 0)
            printf(" * \t%s%s%s\n", ConvertedName, Tabs, Status);
         else
            printf("\t%s%s%s\n", ConvertedName, Tabs, Status);
         KeyIndex++;
         dwSize = sizeof(QMgrName);
         qq++;
      }

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQQMgrsList : RegOpenKeyEx : lRet=%d\n", lRet);
}

/*
 * Function Name
 *  GetMQQMgrStatus
 *
 * Description
 *  This function will output the names of all created MQ queue managers
 *  and their current status.
 *
 * Input parameters
 *  char *QMgrName - Name of the queue manager
 *  char *Status   - Buffer to insert the status
 *  int dwSize     - size of the status buffer
 *
 * Output
 *  char *Status   - Buffer to insert the status
 *
 * Return Value
 *  None.
 */
void GetMQQMgrStatus(char *QMgrName, char *Status, int dwSize)
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;

   sprintf( lpszSubKey, SUBKEY_MQ_QMGR_STATUS_SETTING, QMgrName );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_QUERY_VALUE,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      lRet = RegQueryValueEx( hkCurKey,
                              "CurrentStatus",
                              NULL,
                              &dwType,
                              Status,
                              &dwSize);

      if (lRet != ERROR_SUCCESS)
         Status[0] = '\0';

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQQMgrStatus : RegOpenKeyEx : lRet=%d\n", lRet);
}

/*
 * Function Name
 *  GetMQDefaultQMgr
 *
 * Description
 *  This function will retrieve the name of the default queue manager, if one is set.
 *
 * Input parameters
 *  char *DefaultQMgrName - Buffer to insert the name of the default QMgr
 *  int dwSize            - size of the status buffer
 *
 * Output
 *  char *DefaultQMgrName - Buffer to insert the name of the default QMgr
 *
 * Return Value
 *  None.
 */
void GetMQDefaultQMgr(char *DefaultQMgrName, int dwSize)
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;
   DWORD dwType = REG_SZ;

   strcpy( lpszSubKey, SUBKEY_MQ_QMGR_DEFAULT_SETTING );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_QUERY_VALUE,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      lRet = RegQueryValueEx( hkCurKey,
                              "Name",
                              NULL,
                              &dwType,
                              DefaultQMgrName,
                              &dwSize);

      if (lRet != ERROR_SUCCESS)
         memset(DefaultQMgrName,'\0', dwSize); // No default - clear field

      RegCloseKey(hkCurKey);
   }
   else
   {
      memset(DefaultQMgrName,'\0', dwSize); // No default - clear field
   }
}

/*
 * Function Name
 *  GetMQIndividualQMgrInfo
 *
 * Description
 *  This function will retrieve various info for each queue manager.
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQIndividualQMgrInfo()
{
   int i = 0;
   char  DefaultQMgrName[MQ_Q_MGR_NAME_LENGTH + 1];
   char  ConvertedName[MQ_Q_MGR_NAME_LENGTH + 1];

   printf("+-----------------------------------------------------+\n");
   printf("| Individual Queue Manager Information                |\n");
   printf("+-----------------------------------------------------+\n");

   GetMQDefaultQMgr(DefaultQMgrName, MQ_Q_MGR_NAME_LENGTH);

   while (QMgrList[i].Name[0] != '\0' )
   {
      ConvertMQObjectName(ConvertedName, QMgrList[i].Name);
      printf("Queue Manager Name: %s\n", ConvertedName);

      if (strcmp(ConvertedName, DefaultQMgrName) == 0)
         printf("  Default = Yes\n");
      else
         printf("  Default = No\n");

      printf("  Listeners:\n");
      GetMQListeners(ConvertedName);

      printf("  Logging:\n");
      GetMQQMgrLog(QMgrList[i].Name);

      printf("  Services:\n");
      GetMQServices(ConvertedName);

      i++;
   }
}

/*
 * Function Name
 *  GetMQListeners
 *
 * Description
 *  This function will retrieve all listeners for a particular queue manager.
 *
 * Input parameters
 *  char *QMgrName - Name of the queue manager
 *
 * Output
 *
 *
 * Return Value
 *  None.
 */
void GetMQListeners(char *qMgrName)
{
   int  more,i;
   OUTPUT_LINE_T listeners[200];
   OUTPUT_LINE_T listenerStatus[10];
   OUTPUT_LINE_T listenerDetails[100];
   OUTPUT_LINE_T *pLL = listeners;
   OUTPUT_LINE_T *pLS = listenerStatus;
   OUTPUT_LINE_T *pLD = listenerDetails;
   char listener[64];
   char lsstatus[100];
   char disListener[100];
   char keyValue[64];
   char ListenerStatusNames[TOTAL_LISTENER_PARMS][16] = {"IP Address", "Port", "Type", "Control"};
   char ListenerStatusKeys[TOTAL_LISTENER_PARMS][16] = {"IPADDR(", "PORT(", "TRPTYPE(", "CONTROL("};

   memset(&listeners, '\0', sizeof(listeners));

   Setup_runmqsc(qMgrName, "dis listener(*)", "LISTENER(", listeners);

   while (pLL->Line[0] != '\0')
   {
      ExtractValue(pLL->Line, "LISTENER(", listener);

      if (listener[0] != '\0')
      {
         printf("    Name = %s\n", listener);

         sprintf(lsstatus, "dis lsstatus('%s') status", listener);
         memset(&listenerStatus, '\0', sizeof(listenerStatus));
         memset(keyValue, '\0', sizeof(keyValue));
         Setup_runmqsc(qMgrName, lsstatus, "(", listenerStatus);
         pLS = listenerStatus;
         more = 1;
         while ( (pLS->Line[0] != '\0') && (more) )
         {
            ExtractValue(pLS->Line, "STATUS(", keyValue);

            if (keyValue[0] != '\0')
               more = 0;
            pLS++;
         }
         printf("    Status = %s\n", keyValue);

         sprintf(disListener, "dis listener('%s') all", listener);
         memset(&listenerDetails, '\0', sizeof(listenerDetails));
         Setup_runmqsc(qMgrName, disListener, "(", listenerDetails);
         for (i=0; i < TOTAL_LISTENER_PARMS; i++)
         {
            pLD = listenerDetails;
            memset(keyValue, '\0', sizeof(keyValue));
            more = 1;
            while ( (pLD->Line[0] != '\0') && (more) )
            {
               ExtractValue(pLD->Line, ListenerStatusKeys[i], keyValue);

               if (keyValue[0] != '\0')
               {
                  printf("    %s = %s\n", ListenerStatusNames[i], keyValue);
                  more = 0;
               }

               pLD++;
            }
         }
      }

      printf("    --------------------\n");
      pLL++;
   }
}

/*
 * Function Name
 *  GetMQQMgrLog
 *
 * Description
 *  This function will retrieve logging values for a particular MQ queue manager.
 *
 * Input parameters
 *  char *QMgrName - Name of the queue manager
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQQMgrLog(char *QMgrName)
{
   DWORD lRet;
   char  lpszSubKey[MAX_PATH+1];
   HKEY  hkCurKey;

   sprintf( lpszSubKey, SUBKEY_MQ_QMGRS_LOG_SETTING, QMgrName );

   lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                        lpszSubKey,
                        0,
                        KEY_QUERY_VALUE,
                        &hkCurKey );

   if (lRet == ERROR_SUCCESS)
   {
      DumpKey( "LogType", hkCurKey, "    ", " ");
      DumpKey( "LogPrimaryFiles", hkCurKey, "    ", " ");
      DumpKey( "LogSecondaryFiles", hkCurKey, "    ", " ");
      DumpKey( "LogPath", hkCurKey, "    ", " ");
      DumpKey( "LogBufferPages", hkCurKey, "    ", " ");
      DumpKey( "LogFilePages", hkCurKey, "    ", " ");

      RegCloseKey(hkCurKey);
   }
   else
      printf("Error: GetMQQMgrLog : RegOpenKeyEx : lRet=%d\n", lRet);

   printf("    --------------------\n");
}

/*
 * Function Name
 *  GetMQServices
 *
 * Description
 *  This function will retrieve all services for a particular queue manager.
 *
 * Input parameters
 *  char *QMgrName - Name of the queue manager
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void GetMQServices(char *qMgrName)
{
   int  more, i;
   OUTPUT_LINE_T services[200];
   OUTPUT_LINE_T serviceStatus[10];
   OUTPUT_LINE_T serviceDetails[100];
   OUTPUT_LINE_T *pLL = services;
   OUTPUT_LINE_T *pLS = serviceStatus;
   OUTPUT_LINE_T *pLD = serviceDetails;
   char service[64];
   char lsstatus[100];
   char disService[100];
   char keyValue[64];
   char ServiceStatusNames[TOTAL_SERVICE_PARMS][16] = {"Control", "Service Type"};
   char ServiceStatusKeys[TOTAL_SERVICE_PARMS][16] = {"CONTROL(", "SERVTYPE("};

   memset(&services, '\0', sizeof(services));

   Setup_runmqsc(qMgrName, "dis service(*)", "SERVICE(", services);

   while (pLL->Line[0] != '\0')
   {
      ExtractValue(pLL->Line, "SERVICE(", service);

      if (service[0] != '\0')
      {
         printf("    Name = %s\n", service);

         sprintf(lsstatus, "dis svstatus('%s') status", service);
         memset(&serviceStatus, '\0', sizeof(serviceStatus));
         memset(keyValue, '\0', sizeof(keyValue));
         Setup_runmqsc(qMgrName, lsstatus, "(", serviceStatus);
         pLS = serviceStatus;
         more = 1;
         while ( (pLS->Line[0] != '\0') && (more) )
         {
            ExtractValue(pLS->Line, "STATUS(", keyValue);

            if (keyValue[0] != '\0')
               more = 0;
            pLS++;
         }
         printf("    Status = %s\n", keyValue);

         sprintf(disService, "dis service('%s') all", service);
         memset(&serviceDetails, '\0', sizeof(serviceDetails));
         Setup_runmqsc(qMgrName, disService, "(", serviceDetails);
         for (i=0; i < TOTAL_SERVICE_PARMS; i++)
         {
            pLD = serviceDetails;
            memset(keyValue, '\0', sizeof(keyValue));
            more = 1;
            while ( (pLD->Line[0] != '\0') && (more) )
            {
               ExtractValue(pLD->Line, ServiceStatusKeys[i], keyValue);

               if (keyValue[0] != '\0')
               {
                  printf("    %s = %s\n", ServiceStatusNames[i], keyValue);
                  more = 0;
               }

               pLD++;
            }
         }
      }

      printf("    --------------------\n");
      pLL++;
   }

   // Output a separate line if no data was returned by runmqsc
   if (services[0].Line[0] == '\0')
      printf("  --------------------\n");
}

/*
 * Function Name
 *  Setup_runmqsc
 *
 * Description
 *  Setup to execute a runmqsc command.
 *
 * Input parameters
 *  char *qMgrName - queue maanger name
 *  char *cmd - command to be passed to runmqsc
 *  char *searchString - search string to be used to select the responses
 *
 * Output
 *  OUTPUT_LINE_T *response - response buffer
 *
 * Return Value
 *  None.
 */
void Setup_runmqsc(char *qMgrName, char *runmqscCMD, char *searchString, OUTPUT_LINE_T *response )
{
   DWORD dwWritten;
   SECURITY_ATTRIBUTES saAttr;

   // Set default values
   hChildStd_IN_Rd = NULL;
   hChildStd_IN_Wr = NULL;
   hChildStd_OUT_Rd = NULL;
   hChildStd_OUT_Wr = NULL;

   // Set the bInheritHandle flag so pipe handles are inherited.
   saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
   saAttr.bInheritHandle = TRUE;
   saAttr.lpSecurityDescriptor = NULL;

   // Create a pipe for the child process's STDOUT.
   if ( ! CreatePipe(&hChildStd_OUT_Rd, &hChildStd_OUT_Wr, &saAttr, 0) )
   {
      printf("Error: StdoutRd CreatePipe\n");
      exit(-1);
   }

   // Ensure the read handle to the pipe for STDOUT is not inherited.
   if ( ! SetHandleInformation(hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0) )
   {
      printf("Error: StdoutRd SetHandleInformation\n");
      exit(-1);
   }

   // Create a pipe for the child process's STDIN.
   if (! CreatePipe(&hChildStd_IN_Rd, &hChildStd_IN_Wr, &saAttr, 0))
   {
      printf("Error: Stdin CreatePipe\n");
      exit(-1);
   }

   // Ensure the write handle to the pipe for STDIN is not inherited.
   if ( ! SetHandleInformation(hChildStd_IN_Wr, HANDLE_FLAG_INHERIT, 0) )
   {
      printf("Error: Stdin SetHandleInformation\n");
      exit(-1);
   }

   // Create the child process.
   Launch_runmqsc(qMgrName);

   // Write to the pipe that is the standard input for a child process.
   // Data is written to the pipe's buffers, so it is not necessary to wait
   // until the child process is running before writing data.
   WriteFile(hChildStd_IN_Wr, runmqscCMD, strlen(runmqscCMD), &dwWritten, NULL);

   // Close the pipe handle so the child process stops reading.
   if ( ! CloseHandle(hChildStd_IN_Wr) )
      printf("Error: StdInWr CloseHandle\n");

   // Read from pipe that is the standard output for child process.
   ReadFromPipe(searchString, response);

   // Close remaining handles.
   if ( ! CloseHandle(hChildStd_IN_Rd) )
      printf("Error: StdInRd CloseHandle\n");
   if ( ! CloseHandle(hChildStd_OUT_Rd) )
      printf("Error: StdOutRd CloseHandle\n");
}

/*
 * Function Name
 *  Launch_runmqsc
 *
 * Description
 *  Create a child process for runmqsc that uses the previously created
 *  pipes for STDIN and STDOUT.
 *
 * Input parameters
 *  char *qMgrName - queue maanger name
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void Launch_runmqsc(char *qMgrName)
{
   TCHAR szCmdline[100];
   PROCESS_INFORMATION piProcInfo;
   STARTUPINFO siStartInfo;
   BOOL bSuccess = FALSE;

   // Set up members of the PROCESS_INFORMATION structure.
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );

   sprintf(szCmdline, "runmqsc -e %s", qMgrName);

   // Set up members of the STARTUPINFO structure.
   // This structure specifies the STDIN and STDOUT handles for redirection.
   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO);
   siStartInfo.hStdError = hChildStd_OUT_Wr;
   siStartInfo.hStdOutput = hChildStd_OUT_Wr;
   siStartInfo.hStdInput = hChildStd_IN_Rd;
   siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

   // Create the child process.

   bSuccess = CreateProcess(NULL,
                           szCmdline,     // command line
                           NULL,          // process security attributes
                           NULL,          // primary thread security attributes
                           TRUE,          // handles are inherited
                           0,             // creation flags
                           NULL,          // use parent's environment
                           NULL,          // use parent's current directory
                           &siStartInfo,  // STARTUPINFO pointer
                           &piProcInfo);  // receives PROCESS_INFORMATION

   // If an error occurs, exit the application.
   if ( ! bSuccess )
   {
      printf("Error: CreateProcess\n");
      exit(-1);
   }
   else
   {
      // Close handles to the child process and its primary thread.
      CloseHandle(piProcInfo.hProcess);
      CloseHandle(piProcInfo.hThread);
   }
}

/*
 * Function Name
 *  ReadFromPipe
 *
 * Description
 *  Read output from the child process's pipe for STDOUT
 *  and write the matching output line to the response buffer.
 *  Stop when there is no more data.
 *
 * Input parameters
 *  char *searchString - Search for the string in the response
 *  OUTPUT_LINE_T *response - pointer to the response buffer
 *
 * Output
 *  OUTPUT_LINE_T *response - response data
 *
 * Return Value
 *  None.
 */
void ReadFromPipe(char *searchString, OUTPUT_LINE_T *response)
{
   DWORD dwRead;
   CHAR chBuf[PIPE_BUFFER_SIZE];
   BOOL bSuccess = FALSE;
   HANDLE hParentStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
   char *ptr;
   int i=0;
   OUTPUT_LINE_T *rr = response;

   memset(chBuf, '\0', PIPE_BUFFER_SIZE);

   // Close the write end of the pipe before reading from the
   // read end of the pipe, to control child process execution.
   // The pipe is assumed to have enough buffer space to hold the
   // data the child process has already written to it.
   if (!CloseHandle(hChildStd_OUT_Wr))
      printf("Error: StdInWr CloseHandle\n");

   for (;;)
   {
      bSuccess = ReadFile( hChildStd_OUT_Rd, chBuf, PIPE_BUFFER_SIZE, &dwRead, NULL);
      if( ! bSuccess || dwRead == 0 ) break;

      ptr = strtok(chBuf, "\n");

      while (ptr != NULL)
      {
         if (strstr(ptr, searchString) != NULL)
         {
//            printf("[%d] ptr=%s\n", i++, ptr);
            strcpy(rr->Line, ptr);
            rr++;
         }
//         else
//            printf(" ********** ptr=%s\n", ptr);

         ptr = strtok(NULL, "\n");
      }

      if (! bSuccess ) break;

      memset(chBuf, '\0', PIPE_BUFFER_SIZE);
   }
}

/*
 * Function Name
 *  DumpKey
 *
 * Description
 *  This function will query a specify registry key (type REG_SZ)
 *  then output its value to the user's screen.
 *
 * Input parameters
 *  char *keyName  - the name of the key
 *  HKEY  hkCurKey - address of name of subkey that was opened
 *  char *tabs     - tabs to be outputted
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void DumpKey(char *keyName, HKEY  hkCurKey, char *prefix, char *suffix)
{
   DWORD lRet;
   DWORD dwType = REG_SZ;
   DWORD dwSize = MAX_PATH;
   char  value[MAX_PATH+1];
   char *x = NULL;

   memset(value, '\0', sizeof(value) );

   lRet = RegQueryValueEx( hkCurKey,
                           keyName,
                           NULL,
                           &dwType,
                           value,
                           &dwSize);

   if (lRet != ERROR_SUCCESS)
      value[0] = '\0';

   x = strchr(keyName, '\\');
   if (x != NULL)
      *x = '\0';

   printf("%s%s%s= %s\n", prefix, keyName, suffix, value);
}

/*
 * Function Name
 *  DumpKey_DWord
 *
 * Description
 *  This function will query a specify registry key (type REG_DWORD)
 *  then output its value to the user's screen.
 *
 * Input parameters
 *  char *keyName  - the name of the key
 *  HKEY  hkCurKey - address of name of subkey that was opened
 *  char *tabs     - tabs to be outputted
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void DumpKey_DWord(char *keyName, HKEY  hkCurKey, char *tabs)
{
   DWORD lRet;
   DWORD dwType = REG_DWORD;
   DWORD iSize = 4;
   int dValue = 0;

   lRet = RegQueryValueEx( hkCurKey,
                           keyName,
                           NULL,
                           &dwType,
                           (LPBYTE)&dValue,
                           &iSize);

   if (lRet != ERROR_SUCCESS)
      dValue = -1;

   printf("%s%s= %d\n",keyName,tabs,dValue);
}

/*
 * Function Name
 *  CreateTabs
 *
 * Description
 *  This function will calculate how many tabs are needed.
 *
 * Input parameters
 *  char *Field - the name of the key
 *  char *Tabs  - buffer to insert the tabs
 *  int count   - max tab count
 *
 * Output
 *  char *Tabs  - buffer to insert the tabs
 *
 * Return Value
 *  None.
 */
void CreateTabs(char *Field, char *Tabs, int TabsSize, int count)
{
   int   tc,i;

   /* Calculate how many tabs we need. */
   tc = (count - (strlen(Field) + 6) / 7);
   memset(Tabs, '\0', TabsSize);
   for (i=0; i < tc; i++)
      Tabs[i] = '\t';

   /*
    * Special case: output name is different from key name
    */
   if (memcmp(Field, "Local Clients", 13) == 0)
      strcpy(Tabs, "\t\t");
}

/*
 * Function Name
 *  ExtractValue
 *
 * Description
 *  This function will calculate how many tabs are needed.
 *
 * Input parameters
 *  char *InBuffer - input buffer to be searched
 *  char *key  - key/token for search
 *
 * Output
 *  char *OutBuffer  - out buffer
 *
 * Return Value
 *  None.
 */
void ExtractValue(char *InBuffer, char *key, char *OutBuffer)
{
   char *lb, *rb, *p;

   OutBuffer[0] = '\0'; // default to nothing

   p = strstr(InBuffer, key);
   if (p != NULL)
   {
      lb = p + strlen(key);
      rb = strchr(lb, ')');
      if (rb != NULL)
         *rb = '\0';

      strcpy(OutBuffer, lb);
   }
}

/*
 * Function Name
 *  ConvertMQObjectName
 *
 * Description
 *  This function will
 *
 * Input parameters
 *  None.
 *
 * Output
 *  None.
 *
 * Return Value
 *  None.
 */
void ConvertMQObjectName(char *ConvertedName, char *ObjectName)
{
   int   i;

   strcpy(ConvertedName, ObjectName);
   for (i=0; i < (int)strlen(ConvertedName); i++)
   {
      if (ConvertedName[i] == '!')
         ConvertedName[i] = '.';
      else if (ConvertedName[i] == '/')
         ConvertedName[i] = '&';
   }
}
